from collections import defaultdict, namedtuple
from typing import DefaultDict

from .manifests import Manifest
from .parser import Parser

Command = namedtuple('Command', ['name', 'script'])
CLI = namedtuple('CLI', ['name', 'version', 'code'])


class Commander:
    """Generates commands based on the command config"""

    __slots__ = ('manifest', 'parser', 'cli', 'groups', 'greedy')

    def __init__(self, manifest: Manifest) -> None:
        self.manifest = manifest
        self.parser = Parser(self.manifest)
        self.cli: str = ""
        self.groups: DefaultDict[str, dict] = defaultdict(lambda: defaultdict())
        self.greedy = []

    def build_cli(self) -> None:
        self.add_base_imports()
        self.add_imports()
        self.add_base_cli()
        self.add_functions()
        for name, script in self.manifest.commands.items():
            current_command = Command(name, script)
            self.add_command(current_command)

        self.add_greedy_commands()

    def add_command(self, command: Command) -> None:
        # Check for greedy commands- evaluate them at the end
        if self.is_greedy(command.name):
            self.greedy.append(command)
            return

        if '.' in command.name:
            # Sub command- nested commands
            group = command.name.split('.')[:-1][-1]

            if group not in self.groups:
                self.add_group(group, command)

            self.groups[group][command.name] = command
            self.add_sub_command(command, group)
        else:
            # Group command- top-level commands
            if command.name not in self.groups:
                self.groups[command.name] = {}

                # TODO: by default, add a group app from here to allow for group-level invokes
                # self.add_group(command.name, command)

            self.add_group_command(command)

    def add_imports(self) -> None:
        if isinstance(self.manifest.imports, str):
            self.cli += self.manifest.imports + "\n"
        elif isinstance(self.manifest.imports, list):
            for _import in self.manifest.imports:
                self.cli += _import + "\n"

    def add_functions(self) -> None:
        self.cli += "\n"
        for func in self.manifest.functions:
            self.cli += f"{func}"
        self.cli += "\n"

    def add_greedy_commands(self) -> None:
        """Greedy commands get lazy-loaded. Only supported for group-commands currently"""
        for greedy_command in self.greedy:
            if greedy_command.name.startswith('(*)'):
                for group in self.groups:
                    # make it lazy and interpolate
                    lazy_command_name = greedy_command.name.replace('(*)', group)
                    lazy_command_script = greedy_command.script.replace('{(*)}', group)

                    # lazy load the greedy args
                    greedy_command_args = self.manifest.args.get(greedy_command.name)
                    if greedy_command_args:
                        self.manifest.args[lazy_command_name] = greedy_command_args

                    # lazy parse
                    self.add_command(Command(lazy_command_name, lazy_command_script))

    def is_greedy(self, val: str) -> bool:
        """Greedy strings must contain (*)- marked to be evaluated lazily."""
        return '(*)' in val

    def add_group(self, group: str, command: Command) -> None:
        raise NotImplementedError

    def add_base_imports(self) -> None:
        raise NotImplementedError

    def add_base_cli(self) -> None:
        raise NotImplementedError

    def add_group_command(self, command: Command) -> None:
        raise NotImplementedError

    def add_sub_command(self, command: Command, group: str) -> None:
        raise NotImplementedError


def build_cli(manifest: Manifest, commander_cls=Commander) -> CLI:
    commander = commander_cls(manifest)
    commander.build_cli()
    return CLI(manifest.name, manifest.version, commander.cli)



import datetime

from ..commander import Command, Commander


class TyperCommander(Commander):
    """Generates commands based on the command config"""

    def add_base_imports(self):
        self.cli = f"""## Generated {self.manifest.name} on {datetime.datetime.now()}
import typer; import subprocess; from typing import Optional;
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help']);
"""

    def add_base_cli(self) -> None:
        self.cli += """
cli = typer.Typer(context_settings=CONTEXT_SETTINGS"""

        if self.manifest.cli_options:
            self.cli += f",{self.parser.to_args(self.manifest.cli_options)}"
        if self.manifest.help:
            self.cli += f', help="{self.manifest.help}"'

        self.cli += f""")
__version__ = '{self.manifest.version}'
__cli_name__ = '{self.manifest.name}'

def version_callback(value: bool):
    if value:
        print(__cli_name__ + ", " + __version__)
        raise typer.Exit()

@cli.callback()
def main(version: Optional[bool] = typer.Option(None, '--version', callback=version_callback, is_eager=True)):
    pass
"""

    def add_group(self, group: str, command: Command) -> None:
        self.cli += f"""{group}_app = typer.Typer(); cli.add_typer({group}_app, name="{group}");"""

    def add_group_command(self, command: Command) -> None:
        self.cli += f"""
@cli.command("{self.parser.get_parsed_command_name(command)}")
def {self.parser.get_command_func_name(command)}({self.parser.parse_args(command)}):
{self.parser.parse_command(command.script)}

"""

    def add_sub_command(self, command: Command, group: str) -> None:
        self.cli += f"""
@{group}_app.command("{self.parser.get_parsed_command_name(command)}")
def {self.parser.get_command_func_name(command)}({self.parser.parse_args(command)}):
{self.parser.parse_command(command.script)}

"""



## CLI to generate CLIs
from typing import TextIO

import rich_click as click
from rich.console import Console
from rich.syntax import Syntax
from rich_click.rich_group import RichGroup

from .helper import print_rich_table, write_to_file
from .homer import Homer
from .loader import Loader
from .manifests import Manifest, set_manifest_version
from .transformer import Transformer

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


class RichAliasedGroup(RichGroup):
    def get_command(self, ctx, cmd_name):
        try:
            cmd_name = ALIASES[cmd_name].name
        except KeyError:
            pass
        return super().get_command(ctx, cmd_name or "")


@click.group(context_settings=CONTEXT_SETTINGS, cls=RichAliasedGroup)
@click.version_option()
def cli() -> None:
    pass


@cli.command()
@click.argument('manifests', type=click.File('rb'), nargs=-1)
def load(manifests: list[TextIO]) -> None:
    """Load CLI for given manifest(s)"""
    for manifest in manifests:
        T = Transformer(manifest)
        Loader.load_cli(T.cli)
        Homer.save_cli_metadata(manifest.name, T.cli)
        click.secho(f"~ Generated {T.cli.name} CLI v{T.cli.version} ~", fg="green")
        click.secho(click.style("$", fg="magenta"), nl=False)
        click.echo(f" {T.cli.name} -h")


@cli.command()
@click.argument('cli_names', type=str, nargs=-1)
def update(cli_names: list[str]) -> None:
    """Reloads CLI by name"""
    for cli_name in cli_names:
        cli_metadata = Homer.get_cli_metadata(cli_name)
        if cli_metadata:
            T = Transformer(open(cli_metadata.runner_path, "r"))
            Loader.load_cli(T.cli)
            Homer.save_cli_metadata(cli_metadata.runner_path, T.cli)
            click.secho(f"~ Reloaded {T.cli.name} CLI v{T.cli.version} ~", fg="green")
            click.secho(click.style("$", fg="magenta"), nl=False)
            click.echo(f" {T.cli.name} -h")
        else:
            click.secho(f"~ {cli_name} not found", fg="red")


@cli.command()
@click.argument('manifest', type=click.File('rb'))
def render(manifest: TextIO) -> None:
    """Render the CLI manifest generation as code"""
    T = Transformer(manifest)
    syntax = Syntax(T.cli.code, "python", theme="monokai", line_numbers=False)
    console = Console()
    console.print(syntax)
    click.secho(f"# Rendered {T.cli.name} CLI v{T.cli.version} ~", fg="green")


@cli.command()
@click.argument('cli_name', type=str, default="cliffy")
@click.option('--version', '-v', type=str, show_default=True, default="v1", help="Manifest version")
@click.option(
    '--render', type=bool, is_flag=True, show_default=True, default=False, help="Render template to terminal directly"
)
@click.option(
    '--raw',
    type=bool,
    is_flag=True,
    show_default=True,
    default=False,
    help="Raw template without boilerplate helpers and examples.",
)
def init(cli_name: str, version: str, render: bool, raw: bool) -> None:
    """Generate a CLI manifest template"""
    set_manifest_version(version)
    template = Manifest.get_raw_template(cli_name) if raw else Manifest.get_template(cli_name)

    if render:
        syntax = Syntax(template, "yaml", theme="monokai", line_numbers=False)
        console = Console()
        console.print(syntax)
    else:
        write_to_file(f'{cli_name}.yaml', text=template)
        click.secho(f"+ {cli_name}.yaml", fg="green")


@cli.command("list")
def list_clis() -> None:
    "List all CLIs loaded"
    cols = ["Name", "Version", "Manifest"]
    rows = []
    for metadata in Homer.get_clis():
        rows.append([metadata.cli_name, metadata.version, metadata.runner_path])

    print_rich_table(cols, rows, styles=["cyan", "magenta", "green"])


@cli.command()
@click.argument('cli_names', type=str, nargs=-1)
def remove(cli_names: list[str]) -> None:
    "Remove a loaded CLI by name"
    for cli_name in cli_names:
        if Homer.get_cliffy_cli(cli_name):
            Homer.remove_cli_metadata(cli_name)
            Loader.unload_cli(cli_name)
            click.secho(f"~ {cli_name} unloaded", fg="green")
        else:
            click.secho(f"~ {cli_name} not found", fg="red")


@cli.command()
@click.argument('cli_name', type=str)
def enable(cli_name) -> None:
    click.echo("# TODO")


@cli.command()
@click.argument('cli_name', type=str)
def disable(cli_name) -> None:
    click.echo("# TODO")


ALIASES = {
    "rm": remove,
    "ls": list_clis,
}



from . import v1

Manifest = v1.CLIManifest
IncludeManifest = v1.IncludeManifest


def set_manifest_version(manifestVersion: str) -> None:
    global Manifest
    global IncludeManifest
    if manifestVersion == "v1":
        Manifest = v1.CLIManifest
        IncludeManifest = v1.IncludeManifest



import glob
import json
import os
from base64 import b32encode
from datetime import datetime
from pathlib import Path
from typing import Iterator, Optional

from .commander import CLI
from .helper import CLIFFY_HOME_PATH, write_to_file
from .manifests.v1 import CLIMetadata


class Homer:
    @staticmethod
    def save_cli_metadata(manifest_path: str, cli: CLI) -> None:
        """Stores CLI metadata

        Args:
            manifest_path (str): CLI manifest path
            cli (CLI): CLI
        """
        abs_manifest_path = os.path.realpath(manifest_path)
        encoded_runnerpath = b32encode(cli.name.encode('ascii')).decode('utf-8')
        save_metadata_path = f"{CLIFFY_HOME_PATH}/{encoded_runnerpath}/{cli.name}.json"
        with open(manifest_path, "r") as manifest:
            write_to_file(
                save_metadata_path,
                json.dumps(
                    CLIMetadata(
                        cli_name=cli.name,
                        runner_path=abs_manifest_path,
                        version=cli.version,
                        loaded=datetime.now(),
                        manifest=manifest.read(),
                    ).dict(),
                    default=str,
                ),
            )

    @staticmethod
    def remove_cli_metadata(cli_name: str) -> None:
        """Clears CLI metadata by name

        Args:
            cli_name (str): CLI name
        """
        metadata_paths = glob.glob(f"{CLIFFY_HOME_PATH}/*/{cli_name}.*")
        for path in metadata_paths:
            parent_path = Path(path).parent.resolve()
            os.remove(path)
            os.rmdir(parent_path)

    @staticmethod
    def get_cli_metadata_bypath(path: Path) -> CLIMetadata:
        """Fetches CLI metadata by metadata path

        Args:
            path (Path): Metadata path

        Returns:
            CLIMetadata: CLI metadata
        """
        try:
            return CLIMetadata(**json.load(path.open("r")))
        except Exception as e:
            return CLIMetadata(
                cli_name=path.name,
                runner_path=path.absolute().name,
                version="error",
                loaded=datetime.now(),
                manifest=f"could not load: {e}",
            )

    @classmethod
    def get_cli_metadata(cls, cli_name: str) -> Optional[CLIMetadata]:
        """Fetches CLI metadata by name

        Args:
            cli_name (str): CLI name

        Returns:
            Optional[CLIMetadata]: CLI metadata
        """
        cli_path = cls.get_cliffy_cli(cli_name)
        if not cli_path:
            return None

        path = Path(cli_path)
        return cls.get_cli_metadata_bypath(path)

    @staticmethod
    def get_cliffy_cli(cli_name: str) -> Optional[str]:
        """Fetches CLI metadata path

        Args:
            cli_name (str): CLI name

        Returns:
            Optional[str]: CLI metadata path
        """
        cli_metadata_path = glob.glob(f"{CLIFFY_HOME_PATH}/*/{cli_name}.json")
        if cli_metadata_path:
            return cli_metadata_path[0]
        return None

    @classmethod
    def get_clis(cls) -> Iterator[CLIMetadata]:
        """Fetches loaded CLIs metadata iteratively

        Yields:
            Iterator[CLIMetadata]: CLI metadata
        """
        metadata_paths = Path(CLIFFY_HOME_PATH).glob('*/*')
        for metadata_path in metadata_paths:
            yield cls.get_cli_metadata_bypath(metadata_path)



import os
from typing import Literal, TextIO

import yaml
from jinja2 import BaseLoader, Environment, FileSystemLoader, Undefined
from typing_extensions import Self

from .commander import build_cli
from .commanders.typer import TyperCommander
from .manifests import IncludeManifest, Manifest, set_manifest_version
from .merger import cliffy_merger


class Transformer:
    """Loads command manifest and transforms it into a CLI"""

    def __init__(self, manifest_io: TextIO, as_include: bool = False) -> None:
        self.manifest_io = manifest_io
        self.command_config = self.load_manifest(manifest_io)
        self.manifest_version = self.command_config.pop('manifestVersion', 'v1')

        if self.command_config.get("includes"):
            self.includes_config = self.resolve_includes()
            cliffy_merger.merge(self.command_config, self.includes_config)

        set_manifest_version(self.manifest_version)
        if as_include:
            self.manifest = IncludeManifest(**self.command_config)
        else:
            self.manifest = Manifest(**self.command_config)
            self.cli = build_cli(self.manifest, commander_cls=TyperCommander)

    def resolve_includes(self) -> dict:
        include_transforms = map(self.resolve_include_by_path, set(self.command_config['includes']))
        merged_config = {}
        for transformed_include in include_transforms:
            cliffy_merger.merge(merged_config, transformed_include.command_config)

        return merged_config

    @classmethod
    def resolve_include_by_path(cls, path) -> Self:
        with open(path, "r") as m:
            return cls(m, as_include=True)

    @staticmethod
    def load_manifest(manifest_io: TextIO) -> dict:
        try:
            manifest_path = os.path.realpath(manifest_io.name)
            vars = yaml.safe_load(open(manifest_path, "r")).get('vars', {})
            var_env = Environment(loader=BaseLoader())
            interpolated_vars = {
                var_env.from_string(str(k)).render(vars): var_env.from_string(str(v)).render(vars)
                for k, v in vars.items()
            }
            manifest_env = Environment(loader=FileSystemLoader(manifest_path)).get_template("")
            return yaml.safe_load(manifest_env.render(interpolated_vars))
        except yaml.YAMLError as e:
            print("load_manifest", e)
            raise


class NullUndefined(Undefined):
    def __getattr__(self, key) -> Literal['']:
        return ''



import os
import sys
from pathlib import Path
from typing import Optional

from rich.console import Console
from rich.table import Table

HOME_PATH = str(Path.home())
PYTHON_BIN = f"{sys.exec_prefix}/bin"
PYTHON_EXECUTABLE = sys.executable
CLIFFY_CLI_DIR = f"{Path(__file__).parent.resolve()}/clis"
CLIFFY_HOME_PATH = f"{HOME_PATH}/.cliffy"


def write_to_file(path: str, text: str, executable: bool = False) -> bool:
    try:
        output_file = Path(path)
        output_file.parent.mkdir(exist_ok=True, parents=True)
        output_file.write_text(text)
    except Exception as e:
        print("write_to_file", e)
        return False

    if executable:
        make_executable(path)

    return True


def make_executable(path: str) -> None:
    mode = os.stat(path).st_mode
    mode |= (mode & 0o444) >> 2
    os.chmod(path, mode)


def wrap_as_comment(text: str, split_on: Optional[str] = None) -> str:
    if split_on:
        joiner = "\n# "
        return "# " + joiner.join(text.split(split_on))

    return f"# {text}"


def wrap_as_var(text: str) -> str:
    return '{{' + text + '}}'


def print_rich_table(cols: list[str], rows: list[list[str]], styles: list[str]) -> None:
    table = Table()
    for style, col in zip(styles, cols):
        table.add_column(col, style=style)
    for row in rows:
        table.add_row(*row)

    console = Console()
    console.print(table)
